@@ -28,6 +28,7 @@ import java.util.List;
 import java.util.Map;
 import java.util.Objects;
 import java.util.Optional;
+import java.util.TreeMap;
 import java.util.function.Consumer;
 import java.util.stream.Stream;
 
@@ -41,9 +42,7 @@ public class ServiceDiscoveryImpl {
 
     private static final String CONF_FILE_PATH =  "META-INF/" + CONF_FILE_NAME;
 
-//    private ClassLoader classloader;
-
-    private ServiceDiscoveryImpl() {}
+    ServiceDiscoveryImpl() {}
 
     private static class LazyHolder {
         static final ServiceDiscoveryImpl INSTANCE = new ServiceDiscoveryImpl();
@@ -53,19 +52,11 @@ public class ServiceDiscoveryImpl {
         return LazyHolder.INSTANCE;
     }
 
-    private Map<String, Object>     services   = new HashMap<>();
-    private Map<String, List<?>>    childServices = new HashMap<>();
-    private boolean                 sealed     = false;
-    private boolean                 kiecConfDiscoveryAllowed = true;
-    private Map<String, Object>     cachedServices = new HashMap<String, Object>();
+    private final PriorityMap<String, Object> services = new PriorityMap<>();
+    private final Map<String, List<?>> childServices = new HashMap<>();
 
-    public synchronized boolean isKiecConfDiscoveryAllowed() {
-        return kiecConfDiscoveryAllowed;
-    }
-
-    public synchronized void setKiecConfDiscoveryAllowed(boolean kiecConfDiscoveryAllowed) {
-        this.kiecConfDiscoveryAllowed = kiecConfDiscoveryAllowed;
-    }
+    private Map<String, List<Object>> cachedServices;
+    private boolean sealed = false;
 
     public <T> void addService(Class<T> serviceClass, T service) {
         addService( serviceClass.getCanonicalName(), service );
@@ -73,36 +64,34 @@ public class ServiceDiscoveryImpl {
 
     public synchronized void addService(String serviceName, Object object) {
         if (!sealed) {
-            cachedServices.put(serviceName, object);
+            cachedServices.computeIfAbsent(serviceName, n -> new ArrayList<>()).add(object);
         } else {
             throw new IllegalStateException("Unable to add service '" + serviceName + "'. Services cannot be added once the ServiceDiscoverys is sealed");
         }
     }
 
     public synchronized void reset() {
-        cachedServices = new HashMap<String, Object>();
+        cachedServices = null;
         sealed = false;
+        services.reset();
     }
 
-    public synchronized Map<String, Object> getServices() {
+    public synchronized Map<String, List<Object>> getServices() {
         if (!sealed) {
-            if (kiecConfDiscoveryAllowed) {
-                getKieConfs().ifPresent( kieConfs -> {
-                    while (kieConfs.resources.hasMoreElements()) {
-                        registerConfs( kieConfs.classLoader, kieConfs.resources.nextElement() );
-                    }
-                } );
-                buildMap();
-            }
+            getKieConfs().ifPresent( kieConfs -> {
+                while (kieConfs.resources.hasMoreElements()) {
+                    registerConfs( kieConfs.classLoader, kieConfs.resources.nextElement() );
+                }
+            } );
 
-            cachedServices = Collections.unmodifiableMap( cachedServices );
+            cachedServices = Collections.unmodifiableMap( buildMap() );
             sealed = true;
         }
         return cachedServices;
     }
 
     public void registerConfs( ClassLoader classLoader, URL url ) {
-        log.debug("Loading kie.conf from  " + url + " in classloader " + classLoader);
+        log.debug("Loading kie.conf from {} in classloader {}", url, classLoader);
 
         try ( BufferedReader br = new BufferedReader(new InputStreamReader(url.openStream())) ) {
             for (String line = br.readLine(); line != null; line = br.readLine()) {
@@ -112,8 +101,8 @@ public class ServiceDiscoveryImpl {
                     processKieService( classLoader, entry[0].trim(), entry[1].trim() );
                 }
             }
-        } catch (Exception exc) {
-            throw new RuntimeException( "Unable to build kie service url = " + url.toExternalForm(), exc );
+        } catch (Exception e) {
+            throw new RuntimeException( "Unable to build kie service url = " + url.toExternalForm(), e );
         }
     }
 
@@ -125,34 +114,67 @@ public class ServiceDiscoveryImpl {
                 if ( value.startsWith( "+" ) ) {
                     childServices.computeIfAbsent( serviceName, k -> new ArrayList<>() )
                             .add( newInstance( classLoader, value.substring( 1 ) ) );
+                    log.debug("Added child Service {}", value );
                 } else {
-                    services.put( serviceName, newInstance( classLoader, value ) );
+                    String[] splitValues = value.split( ";" );
+                    if (splitValues.length > 2) {
+                        throw new RuntimeException( "Invalid kie.conf entry: " + value );
+                    }
+                    int priority = splitValues.length == 2 ? Integer.parseInt( splitValues[1].trim() ) : 0;
+                    services.put( priority, serviceName, newInstance( classLoader, splitValues[0].trim() ) );
+                    log.debug( "Added Service {} with priority {}", value, priority );
                 }
             } catch (RuntimeException e) {
                 if (optional) {
-                    log.info("Cannot load service: " + serviceName);
+                    log.info("Cannot load service: {}",serviceName);
                 } else {
-                    System.out.println("Loading failed because " + e.getMessage());
+                    log.error("Loading failed because {}", e.getMessage());
                     throw e;
                 }
             }
-            log.debug( "Adding Service {}\n", value );
         }
     }
 
-    @FunctionalInterface
-    private interface ServiceProcessor {
-        boolean process(ClassLoader classLoader, String key, String value);
-    }
-
     private <T> T newInstance( ClassLoader classLoader, String className ) {
         try {
-            return (T) Class.forName( className, true, classLoader ).newInstance();
+            return (T) Class.forName( className, true, classLoader ).getConstructor().newInstance();
         } catch (Throwable t) {
             throw new RuntimeException( "Cannot create instance of class: " + className, t );
         }
     }
 
+    private Map<String, List<Object>> buildMap() {
+        Map<String, List<Object>> servicesMap = new HashMap<>();
+        for (Map.Entry<String, List<Object>> serviceEntry : services.entrySet()) {
+            log.debug( "Service {} is implemented by {}",  serviceEntry.getKey(), serviceEntry.getValue().get(0) );
+            servicesMap.put(serviceEntry.getKey(), serviceEntry.getValue());
+            List<?> children = childServices.remove( serviceEntry.getKey() );
+            if (children != null) {
+                for (Object child : children) {
+                    for (Object service : serviceEntry.getValue()) {
+                        (( Consumer ) service).accept( child );
+                    }
+                }
+            }
+        }
+
+        if (!childServices.isEmpty()) {
+            throw new RuntimeException("Child services " + childServices.keySet() + " have no parent");
+        }
+
+        if (log.isTraceEnabled()) {
+            for (Map.Entry<String, List<Object>> serviceEntry : servicesMap.entrySet()) {
+                if (serviceEntry.getValue().size() == 1) {
+                    log.trace( "Service {} is implemented by {}",  serviceEntry.getKey(), serviceEntry.getValue().get(0) );
+                } else {
+                    log.trace( "Service {} is implemented (in order of priority) by {}", serviceEntry.getKey(), serviceEntry.getValue() );
+                }
+            }
+        }
+
+        return servicesMap;
+    }
+
     private Optional<KieConfs> getKieConfs() {
         return Stream.of(this.getClass().getClassLoader(), Thread.currentThread().getContextClassLoader(), ClassLoader.getSystemClassLoader())
                 .map(this::loadKieConfs)
@@ -182,19 +204,37 @@ public class ServiceDiscoveryImpl {
         }
     }
 
-    private void buildMap() {
-        for (Map.Entry<String, Object> serviceEntry : services.entrySet()) {
-            cachedServices.put(serviceEntry.getKey(), serviceEntry.getValue());
-            List<?> children = childServices.remove( serviceEntry.getKey() );
-            if (children != null) {
-                for (Object child : children) {
-                    ( (Consumer) serviceEntry.getValue() ).accept( child );
-                }
-            }
+    private static class PriorityMap<K,V> {
+        private final Map<K, TreeMap<Integer, V>> priorityMap = new HashMap<>();
+
+
+        public void reset() {
+            priorityMap.clear();
         }
 
-        if (!childServices.isEmpty()) {
-            throw new RuntimeException("Child services " + childServices.keySet() + " have no parent");
+        public void put(int priority, K key, V value) {
+            TreeMap<Integer, V> treeMap = priorityMap.get(key);
+            if ( treeMap == null ) {
+                treeMap = new TreeMap<>();
+                priorityMap.put( key, treeMap );
+            } else {
+                if ( treeMap.get( priority ) != null ) {
+                    throw new RuntimeException("There already exists an implementation for service " + key + " with same priority " + priority);
+                }
+            }
+            treeMap.put( priority, value );
+        }
+
+        public Iterable<? extends Map.Entry<K, List<V>>> entrySet() {
+            Map<K, List<V>> map = new HashMap<>();
+            for (Map.Entry<K, TreeMap<Integer, V>> entry : priorityMap.entrySet()) {
+                List<V> list = new ArrayList<>();
+                for (V value : entry.getValue().values()) {
+                    list.add(0, value);
+                }
+                map.put( entry.getKey(), list );
+            }
+            return map.entrySet();
         }
     }
 }
